import re
from collections.abc import Generator
from os import PathLike
from typing import Any

import frontmatter

from .base_objects import PREFIXES, BaseObj, Cookware, Ingredient
from .const import METADATA_DISPLAY_MAP, METADATA_MAPPINGS


class Metadata:
    """Recipe Metadata Class"""

    def __init__(self, metadata: dict):
        """
        Initialize the Metadata class

        :param metadata: Dictionary of metadata
        """
        self._parsed = {k.strip(): v for k, v in metadata.items()}
        self._mapped = {METADATA_MAPPINGS.get(k.lower(), k.lower()): v for k, v in self._parsed.items()}
        for attr, value in self._mapped.items():
            setattr(self, attr, value)

        for attr, value in self._parsed.items():
            setattr(self, attr, value)

    def __str__(self) -> str:
        s = ''
        for k, v in self._mapped.items():
            s += f'{METADATA_DISPLAY_MAP.get(k, k).capitalize()}: {v}\n'
        if s:
            return s + ('-' * 50) + '\n'
        return s

    def __getitem__(self, key: str) -> Any:
        if key in self._parsed:
            return self._parsed[key]
        if key in self._mapped:
            return self._mapped[key]
        raise KeyError(f'Unable to find {key=}.')

    def get(self, key: str, default: Any = None) -> Any:
        """
        Attempt to get the value associated with key

        :param key: The key to get
        :param default: Default value to return if the key is not found
        :return: The entry at key or default
        """
        try:
            return self[key]
        except KeyError:
            return default


class Recipe:
    def __init__(self, recipe: str, prefixes: dict = PREFIXES):
        """
        Parse the recipe string into a Recipe object.

        :param recipe: Recipe string
        :param prefixes: Prefixes for parsing. Default is PREFIXES constant.
                         This allows for overriding the handling of one or
                         more of the base objects.
        """
        self._raw = recipe
        metadata, body = frontmatter.parse(re.sub(r':(?=\S)', ': ', recipe))
        self.metadata = Metadata(metadata)
        if not body:
            raise ValueError('No body found in recipe!')
        self.steps = list()
        self.ingredients = list()
        self.cookware = list()
        for line in re.split(r'\n{2,}', body):
            line = re.sub(r'\s+', ' ', line)
            if step := Step(line, prefixes):
                self.steps.append(step)
                self.ingredients.extend(step.ingredients)
                self.cookware.extend(step.cookware)

    def __iter__(self) -> Generator:
        yield from self.steps

    def __len__(self) -> int:
        return len(self.steps)

    def __str__(self) -> str:
        s = str(self.metadata)
        s += 'Ingredients:\n\n'
        s += '\n'.join(
            (f'{ing:%q[%af %us] %n (%c)}' if ing.notes else f'{ing:%q[%af %us] %n}') for ing in self.ingredients
        )
        s += '\n' + ('-' * 50) + '\n'
        if self.cookware:
            s += '\nCookware:\n\n'
            s += '\n'.join(
                (f'{cw:%q[%af %us] %n (%c)}' if cw.notes else f'{cw:%q[%af %us] %n}') for cw in self.cookware
            )
            s += '\n' + ('-' * 50) + '\n'
        s += '\n'
        s += '\n'.join(map(str, self))
        return s.replace('\\', '') + '\n'

    @staticmethod
    def from_file(filename: PathLike, prefixes: dict = PREFIXES):
        """
        Load a recipe from a file

        :param filename: Path like object indicating the location of the file.
        :param prefixes: Prefixes for parsing. Default is PREFIXES constant.
                         This allows for overriding the handling of one or
                         more of the base objects.
        :return: Recipe object
        """
        with open(filename) as f:
            return Recipe(f.read(), prefixes)


class Step:
    def __init__(self, line: str, prefixes: dict = PREFIXES):
        """
        Parse a line into its sections and objects

        :param line: The line to parse
        :param prefixes: Prefixes for parsing. Default is PREFIXES constant.
                 This allows for overriding the handling of one or
                 more of the base objects.

        """
        self._raw: str = line
        self.ingredients: list[Ingredient] = list()
        self.cookware: list[Cookware] = list()
        self._sections: list[str | BaseObj] = list()
        self._prefixes: dict[str, type[BaseObj]] = prefixes
        self._parse(line)

    def __iter__(self):
        yield from self._sections

    def __len__(self):
        return len(self._sections)

    def __repr__(self):
        return repr(self._sections)

    def _parse(self, line: str):
        """
        Parse a line into its component parts

        :param line: The line to parse
        """
        if not (section := self._remove_comments(line)):
            return
        self._sections.clear()
        while match := re.search(r'(?<!\\)[@#~][\S]', section):
            if section[: match.start()].strip():
                self._sections.append(section[: match.start()])
            section = section[match.start() :]
            obj = self._prefixes[section[0]].factory(section)
            self._sections.append(obj)
            section = section.removeprefix(obj.raw)
            match obj:
                case Ingredient():
                    self.ingredients.append(obj)
                case Cookware():
                    self.cookware.append(obj)
        if section.strip():
            self._sections.append(section)

    def __str__(self):
        return ''.join(map(str, self)).rstrip()

    @staticmethod
    def _remove_comments(line: str) -> str:
        """
        Remove comments from the line.

        :param line: The line to clean of comments
        :return: The line without comments.
        """
        return re.sub(r'--.*(?:$|\n)|\[-.*?-]', '', line)



"""Utility classes and functions"""

from fractions import Fraction


class WholeFraction(Fraction):
    def __str__(self):
        whole = abs(self.numerator) // self.denominator
        part = abs(self) - whole
        sign = '-' if self < 0 else ''
        if whole and part:
            return f'{sign}{whole} {part}'
        if whole:
            return f'{sign}{whole}'
        if part:
            return f'{sign}{part}'
        return '0'



from .base_objects import BaseObj, Cookware, Ingredient, Quantity, Timing
from .recipe import PREFIXES, Metadata, Recipe, Step
from .utils import WholeFraction

__all__ = [
    'Recipe',
    'Step',
    'Ingredient',
    'Cookware',
    'Timing',
    'Quantity',
    'Metadata',
    'PREFIXES',
    'BaseObj',
    'WholeFraction',
]



"""Base Object for Ingredient, Cookware, and Timing"""

import re
from decimal import Decimal, InvalidOperation

from .const import LONG_TO_SHORT_MAPPINGS, NOTE_PATTERN, QUANTITY_PATTERN, SHORT_TO_LONG_MAPPINGS
from .utils import WholeFraction


class Quantity:
    """Quantity Class"""

    def __init__(self, qstr: str):
        """
        Constructor for the Quantity class

        :param qstr: The quantity string
        """
        self._raw = qstr
        self.unit = ''
        if '%' in qstr:
            self.amount, self.unit = map(str.strip, qstr.split('%'))
        else:
            self.amount = qstr
        self.amount = self.amount.strip()

        # Try storing the quantity as a numeric value
        try:
            if match := re.match(r'(\d+)?\s*(\d+)\s*/\s*(\d+)', self.amount):
                whole, *parts = match.groups()
                whole = int(whole) if whole else 0
                parts = WholeFraction('/'.join(parts))
                self.amount = WholeFraction(whole + parts)
            elif '.' in self.amount:
                self.amount = Decimal(self.amount)
            else:
                self.amount = int(self.amount)
        except (ValueError, InvalidOperation):
            pass

    def __eq__(self, other) -> bool:
        if not isinstance(other, Quantity):
            return False
        return self.amount == other.amount and self.unit == other.unit

    def __str__(self) -> str:
        return f'{self:%a %us}'.strip()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(qstr={repr(self._raw)})'

    def __hash__(self) -> int:
        return hash((self.amount, self.unit))

    def __format__(self, format_spec: str) -> str:
        """
        Return the quantity based on the format spec

        %a - Amount
        %af - Amount as fraction
        %u - Unit
        %ul - Long unit
        %us - short unit
        """
        if not format_spec:
            return str(self)
        s = ''
        fs = iter(format_spec)
        c = next(fs)
        spaces = 0
        while True:
            if c == '%':
                try:
                    c = next(fs)
                except StopIteration:
                    return s
                match c:
                    case 'a':
                        try:
                            c = next(fs)
                        except StopIteration:
                            return s + str(self.amount) if self.amount else ''
                        match c:
                            case 'f':
                                try:
                                    f = WholeFraction(self.amount) if self.amount else ''
                                    s += str(f)
                                    spaces = 0
                                except ValueError:
                                    s += str(self.amount if self.amount else '')
                                    spaces = 0
                            case '%':
                                s += str(self.amount if self.amount else '')
                                spaces = 0
                                continue
                            case _:
                                s += str(self.amount) + c
                                spaces = 0
                    case 'u':
                        try:
                            c = next(fs)
                        except StopIteration:
                            return s + self.unit if self.unit else ''
                        if not self.unit and spaces:
                            s = s[: spaces * -1]
                        match c:
                            case 's':
                                s += LONG_TO_SHORT_MAPPINGS.get(self.unit, self.unit if self.unit else '')
                            case 'l':
                                s += SHORT_TO_LONG_MAPPINGS.get(self.unit, self.unit if self.unit else '')
                            case '%':
                                s += self.unit if self.unit else ''
                                continue
                            case _:
                                s += self.unit if self.unit else '' + c
                    case _:
                        s += f'%{c}'
            else:
                s += c
                if c == ' ':
                    spaces += 1
                else:
                    spaces = 0
            try:
                c = next(fs)
            except StopIteration:
                return s
            continue
        return s

    def __radd__(self, other) -> str:
        if not isinstance(other, str):
            raise TypeError(f'Cannot add {self} to {other.__class__.__name__}')
        return f'{other}{self}'


class BaseObj:
    """Base Object for Ingredient, Cookware, and Timing"""

    prefix = None
    supports_notes = True

    def __init__(
        self,
        raw: str,
        name: str,
        *,
        quantity: str = None,
        notes: str = None,
    ):
        """
        Constructor for the BaseObj class

        :param raw: The raw string the ingredient came from
        :param name: The name of the ingredient
        :param quantity: The quantity as described in the raw string
        :param notes: Notes from the raw string
        """
        self.raw = raw
        self.name = name.strip()
        self._quantity = quantity.strip() if quantity and quantity.strip() else None
        self.notes = notes
        self._parsed_quantity = Quantity(self._quantity) if self._quantity else ''

    def __eq__(self, other) -> bool:
        if not (isinstance(other, BaseObj)):
            return False
        return all(getattr(self, attr) == getattr(other, attr) for attr in ('name', '_parsed_quantity', 'notes'))

    def __repr__(self) -> str:
        s = f'{self.__class__.__name__}(raw={self.raw!r}, name={self.name!r}, quantity={self._quantity!r}'
        if self.__class__.supports_notes:
            s += f', notes={repr(self.notes)}'
        return s + ')'

    @property
    def quantity(self) -> Quantity | str:
        return self._parsed_quantity

    def __str__(self) -> str:
        """Short version of the formatted string"""
        if self.quantity:
            return f'{self.name} ({self.quantity})'.strip()
        return self.name

    def __hash__(self) -> int:
        return hash(tuple(getattr(self, attr) for attr in ('name', '_parsed_quantity', 'notes')))

    def __format__(self, format_spec: str) -> str:
        """
        Format the string

        %n - Name
        %q - Quantity
        %q[<format>] - Quantity as format
        %c - Notes
        """
        if not format_spec:
            return str(self)
        s = ''
        fs = iter(format_spec)
        c = next(fs)
        while True:
            if c == '%':
                try:
                    c = next(fs)
                except StopIteration:
                    return s
                match c:
                    case 'c':
                        if self.notes:
                            s += self.notes
                    case 'n':
                        s += self.name
                    case 'q':
                        try:
                            c = next(fs)
                        except StopIteration:
                            return s + str(self.quantity)
                        qformat_spec = ''
                        if c == '[':
                            try:
                                while (c := next(fs)) != ']':
                                    qformat_spec += c
                            except StopIteration:
                                return s + f'{self.quantity:{qformat_spec}}' if self.quantity else ''
                        s += (f'{self.quantity:{qformat_spec}}' if self.quantity else '') + (c if c != ']' else '')
                    case _:
                        s += f'%{c}'
            else:
                s += c
            try:
                c = next(fs)
            except StopIteration:
                return s
            continue
        return s

    @classmethod
    def factory(cls, raw: str):
        """
        Factory to create an object

        :param raw: raw string to create from
        :return: An object of cls
        """
        if not cls.prefix:
            raise NotImplementedError(f'{cls.__name__} does not have a prefix set!')
        if not raw.startswith(cls.prefix):
            raise ValueError(f'Raw string does not start with {repr(cls.prefix)}: [{repr(raw[0])}]')
        raw = raw[1:]
        if next_object_starts := [raw.index(prefix) for prefix in PREFIXES if prefix in raw]:
            next_start = min(next_object_starts)
            raw = raw[:next_start]
        note_pattern = NOTE_PATTERN if cls.supports_notes else ''
        if match := re.search(rf'(?P<name>.*?){QUANTITY_PATTERN}{note_pattern}', raw):
            return cls(f'{cls.prefix}{raw[: match.end(match.lastgroup) + 1]}', **match.groupdict())
        if note_pattern and (match := re.search(rf'^(P<name>[\S]+){note_pattern}', raw)):
            return cls(f'{cls.prefix}{raw[: match.end(match.lastgroup) + 1]}', **match.groupdict())
        name = raw.split()[0]
        name = re.sub(r'\W+$', '', name) or name
        return cls(f'{cls.prefix}{name}', name=name)

    def __radd__(self, other) -> str:
        if not isinstance(other, str):
            raise TypeError(f'Cannot add {self} to {other.__class__.__name__}')
        return f'{other}{self}'


class Ingredient(BaseObj):
    """Ingredient"""

    prefix = '@'
    supports_notes = True


class Cookware(BaseObj):
    """Cookware"""

    prefix = '#'
    supports_notes = True


class Timing(BaseObj):
    """Timing"""

    prefix = '~'
    supports_notes = False

    def __str__(self) -> str:
        return f'{self.name.strip()} {str(self.quantity).strip()}'

    @property
    def long_str(self) -> str:
        return str(self)


PREFIXES = {
    '@': Ingredient,
    '#': Cookware,
    '~': Timing,
}
